@@ -3,7 +3,7 @@
 import os, os.path, shlex, subprocess, sys, time, termios, json, getpass, reporting
 from dataclasses import dataclass
 
-import system, osenum, stub, diskutil, osinstall, asahi_firmware, m1n1
+import system, osenum, stub, diskutil, osinstall, asahi_firmware, m1n1, bugs
 from util import *
 
 PART_ALIGN = psize("1MiB")
@@ -18,7 +18,7 @@ MIN_FREE = psize("1GB")
 # 2.5GB stub + 5GB OS + 0.5GB EFI = 8GB, round up to 10GB
 MIN_INSTALL_FREE = psize("10GB")
 
-MIN_MACOS_VERSION = "12.3"
+MIN_MACOS_VERSION = "13.5"
 MIN_MACOS_VERSION_EXPERT = "12.3"
 
 @dataclass
@@ -28,6 +28,7 @@ class IPSW:
     min_iboot: str
     min_sfr: str
     expert_only: bool
+    devices: list
     url: str
 
 @dataclass
@@ -64,39 +65,35 @@ DEVICES = {
     "j414sap":  Device("13.2", False),  # MacBook Pro (14-inch, M2 Pro, 2023)
     "j416cap":  Device("13.2", False),  # MacBook Pro (16-inch, M2 Max, 2023)
     "j416sap":  Device("13.2", False),  # MacBook Pro (16-inch, M2 Pro, 2023)
-    "j473ap":   Device("13.2", True),  # Mac mini (M2, 2023)
-    "j474sap":  Device("13.2", True),  # Mac mini (M2 Pro, 2023)
+    "j473ap":   Device("13.2", False),  # Mac mini (M2, 2023)
+    "j474sap":  Device("13.2", False),  # Mac mini (M2 Pro, 2023)
     "j415ap":   Device("13.4", False),  # MacBook Air (15-inch, M2, 2023)
-    "j475cap":  Device("13.4", True),  # Mac Studio (M2 Max, 2023)
-    "j475dap":  Device("13.4", True),  # Mac Studio (M2 Ultra, 2023)
+    "j475cap":  Device("13.4", False),  # Mac Studio (M2 Max, 2023)
+    "j475dap":  Device("13.4", False),  # Mac Studio (M2 Ultra, 2023)
     "j180dap":  Device("13.4", True),  # Mac Pro (M2 Ultra, 2023)
 }
 
 IPSW_VERSIONS = [
-    # This is the special M2 version, it comes ahead so it isn't the default in expert mode
-    IPSW("12.4",
-         "12.1",
-         "iBoot-7459.121.3",
-         "21.6.81.2.0,0",
-         False,
-         "https://updates.cdn-apple.com/2022SpringFCS/fullrestores/012-17781/F045A95A-44B4-4BA9-8A8A-919ECCA2BB31/UniversalMac_12.4_21F2081_Restore.ipsw"),
     IPSW("12.3.1",
          "12.1",
          "iBoot-7459.101.3",
          "21.5.258.0.0,0",
          False,
+         None,
          "https://updates.cdn-apple.com/2022SpringFCS/fullrestores/002-79219/851BEDF0-19DB-4040-B765-0F4089D1530D/UniversalMac_12.3.1_21E258_Restore.ipsw"),
     IPSW("12.3",
          "12.1",
          "iBoot-7459.101.2",
          "21.5.230.0.0,0",
          False,
+         None,
          "https://updates.cdn-apple.com/2022SpringFCS/fullrestores/071-08757/74A4F2A1-C747-43F9-A22A-C0AD5FB4ECB6/UniversalMac_12.3_21E230_Restore.ipsw"),
     IPSW("13.5",
          "13.0",
          "iBoot-8422.141.2",
          "22.7.74.0.0,0",
          False,
+         None,
          "https://updates.cdn-apple.com/2023SummerFCS/fullrestores/032-69606/D3E05CDF-E105-434C-A4A1-4E3DC7668DD0/UniversalMac_13.5_22G74_Restore.ipsw"),
 ]
 
@@ -206,12 +203,31 @@ class InstallerMain:
         p_progress(f"Using OS '{self.cur_os.label}' ({self.cur_os.sys_volume}) for machine authentication.")
         logging.info(f"Current OS: {self.cur_os.label} / {self.cur_os.sys_volume}")
 
-        if not self.cur_os.admin_users:
+        if not self.cur_os.admin_users or self.cur_os.admin_users == ["_mbsetupuser"] or (
+            self.sysinfo.boot_mode == "macOS" and self.sysinfo.login_user not in self.cur_os.admin_users):
+            # Out of date Preboot partition?
+            p_message(f"Oops, your Preboot volume may be out of date. Fixing that for you...")
+            p = subprocess.run(["diskutil", "apfs", "updatePreboot", self.cur_os.system], check=True,
+                               stdout=subprocess.PIPE,
+                               stderr=subprocess.STDOUT)
+            logging.debug(f"process output: {p.stdout}")
+            self.cur_os.update_admin_users()
+            p_message(f"Preboot volume updated.")
+
+        if not self.cur_os.admin_users or self.cur_os.admin_users == ["_mbsetupuser"]:
             p_error("No admin users found in the primary OS. Cannot continue.")
             p_message("If this is a new or freshly reset machine, you will have to go through macOS")
             p_message("initial user set-up and create an admin user before using this installer.")
             sys.exit(1)
 
+        if self.sysinfo.boot_mode == "macOS" and self.sysinfo.login_user not in self.cur_os.admin_users:
+            p_error(f"Your login user {self.sysinfo.login_user} is not a machine admin.")
+            print()
+            p_info("These users are machine admins: {' '.join(self.cur_os.admin_users)}")
+            print()
+            p_message("You have to be a machine admin to install Asahi Linux.")
+            sys.exit(1)
+
     def get_admin_credentials(self):
         if self.credentials_validated:
             return
@@ -266,7 +282,7 @@ class InstallerMain:
         self.osins = osinstall.OSInstaller(self.dutil, self.data, template)
         self.osins.load_package()
 
-        min_size = STUB_SIZE + self.osins.min_size
+        min_size = STUB_SIZE + (self.osins.min_size if self.expert else self.osins.min_recommended_size)
         print()
         p_message(f"Minimum required space for this OS: {ssize(min_size)}")
 
@@ -286,7 +302,7 @@ class InstallerMain:
         self.osins = osinstall.OSInstaller(self.dutil, self.data, template)
         self.osins.load_package()
 
-        min_size = STUB_SIZE + self.osins.min_size
+        min_size = STUB_SIZE + (self.osins.min_size if self.expert else self.osins.min_recommended_size)
         print()
         p_message(f"Minimum required space for this OS: {ssize(min_size)}")
 
@@ -395,6 +411,7 @@ class InstallerMain:
             return False
 
         self.dutil.remount_rw(self.ins.osi.system)
+        self.ins.repair(self.cur_os)
 
         if upgrade:
             # Note: we get the vars out of the boot.bin in the system volume instead of the
@@ -453,7 +470,8 @@ class InstallerMain:
         device_min = split_ver(self.device.min_ver)
         minver = [ipsw for ipsw in IPSW_VERSIONS
                  if split_ver(ipsw.version) >= max(chip_min, device_min)
-                 and (supported_fw is None or ipsw.version in supported_fw)]
+                 and (supported_fw is None or ipsw.version in supported_fw)
+                 and (ipsw.devices is None or self.sysinfo.device_class in ipsw.devices)]
         avail = [ipsw for ipsw in minver
                  if split_ver(ipsw.min_iboot) <= sys_iboot
                  and split_ver(ipsw.min_macos) <= sys_macos
@@ -606,7 +624,7 @@ class InstallerMain:
         print()
         p_message( "When the system shuts down, follow these steps:")
         print()
-        p_message( "1. Wait 15 seconds for the system to fully shut down.")
+        p_message( "1. Wait 25 seconds for the system to fully shut down.")
         p_message(f"2. Press and {col(BRIGHT, YELLOW)}hold{col()}{col(BRIGHT)} down the power button to power on the system.")
         p_warning( "   * It is important that the system be fully powered off before this step,")
         p_warning( "     and that you press and hold down the button once, not multiple times.")
@@ -618,7 +636,7 @@ class InstallerMain:
         p_plain(   "   * If you are asked to 'Select a volume to recover',")
         p_plain(   "     then choose your normal macOS volume and click Next.")
         p_plain(   "     You may need to authenticate yourself with your macOS credentials.")
-        p_message( "7. Once the 'Asahi Linux installer' screen appears, follow the prompts.")
+        p_message(f"7. Once the '{DISTRO} installer' screen appears, follow the prompts.")
         print()
         p_warning( "If you end up in a bootloop or get a message telling you that macOS needs to")
         p_warning( "be reinstalled, that means you didn't follow the steps above properly.")
@@ -800,14 +818,13 @@ class InstallerMain:
 
     def main(self):
         print()
-        p_message("Welcome to the Asahi Linux installer!")
+        p_message(f"Welcome to the {DISTRO} installer!")
         print()
-        p_message("This installer is in an alpha state, and may not work for everyone.")
-        p_message("It is intended for developers and early adopters who are comfortable")
-        p_message("debugging issues or providing detailed bug reports.")
+        p_message("This installer will guide you through the process of setting up")
+        p_message(f"{DISTRO} on your Mac.")
         print()
         p_message("Please make sure you are familiar with our documentation at:")
-        p_plain( f"  {col(BLUE, BRIGHT)}https://alx.sh/w{col()}")
+        p_plain( f"  {col(BLUE, BRIGHT)}{DISTRO_DOCS}{col()}")
         print()
         p_question("Press enter to continue.")
         self.input()
@@ -826,6 +843,9 @@ class InstallerMain:
         self.sysinfo = system.SystemInfo()
         self.sysinfo.show()
         print()
+
+        bugs.run_checks(self)
+
         self.chip_min_ver = CHIP_MIN_VER.get(self.sysinfo.chip_id, None)
         self.device = DEVICES.get(self.sysinfo.device_class, None)
         if not self.chip_min_ver or not self.device or (self.device.expert_only and not self.expert):
@@ -889,6 +909,7 @@ class InstallerMain:
         oses_upgradable = []
 
         for i, p in enumerate(self.parts):
+            p.index = i
             if p.type in ("Apple_APFS_ISC",):
                 continue
             if p.free:
@@ -958,6 +979,8 @@ class InstallerMain:
                 else:
                     state += " "
                 p_plain(f"    OS: [{state}] {os}")
+                if os.attached_partitions:
+                    p_plain(f"      Extra partitions: {' '.join('#' + str(i.index) for i in os.attached_partitions)}")
                 if os.stub and os.m1n1_ver and os.m1n1_ver != self.m1n1_ver:
                     oses_upgradable.append((p, os))
                 elif os.stub and not (os.bp and os.bp.get("coih", None)):
