@@ -18,6 +18,7 @@ class StubInstaller(PackageInstaller):
         self.copy_idata = []
         self.stub_info = {}
         self.pkg = None
+        self.is_ota = False
 
     def load_ipsw(self, ipsw_info):
         self.install_version = ipsw_info.version.split(maxsplit=1)[0]
@@ -27,7 +28,14 @@ class StubInstaller(PackageInstaller):
         if base:
             url = base + "/" + os.path.split(url)[-1]
 
-        logging.info(f"IPSW URL: {url}")
+        if not url.endswith(".ipsw"):
+            self.is_ota = True
+
+        if self.is_ota:
+            logging.info(f"OTA URL: {url}")
+        else:
+            logging.info(f"IPSW URL: {url}")
+
         if url.startswith("http"):
             p_progress("Downloading macOS OS package info...")
             self.ucache = urlcache.URLCache(url)
@@ -116,6 +124,7 @@ class StubInstaller(PackageInstaller):
         self.sv_path = os.path.join(self.core_services, "SystemVersion.plist")
         self.sv_dis_path = os.path.join(self.core_services, "SystemVersion-disabled.plist")
         self.icon_path = os.path.join(self.osi.system, ".VolumeIcon.icns")
+        self.pb_vgid = os.path.join(self.osi.preboot, self.osi.vgid)
 
     def check_existing_install(self, osi):
         self.osi = osi
@@ -146,30 +155,89 @@ class StubInstaller(PackageInstaller):
         if not os.path.exists(self.iapm_path):
             os.replace(self.iapm_dis_path, self.iapm_path)
 
+    def path(self, path):
+        if not self.is_ota:
+            return path
+
+        if path.startswith("BootabilityBundle"):
+            return os.path.join("AssetData", "Restore", path)
+        else:
+            return os.path.join("AssetData", "boot", path)
+
+    def open(self, path):
+        return self.pkg.open(self.path(path))
+
+    def copy_admin_users(self, cur_os):
+        logging.info("Copying admin users")
+
+        os.makedirs(os.path.join(self.pb_vgid, "var/db"), exist_ok=True)
+        admin_users = os.path.join(cur_os.preboot, cur_os.vgid, "var/db/AdminUserRecoveryInfo.plist")
+        tg_admin_users = os.path.join(self.pb_vgid, "var/db/AdminUserRecoveryInfo.plist")
+        if os.path.exists(tg_admin_users):
+            self.chflags("noschg", tg_admin_users)
+        shutil.copy(admin_users, tg_admin_users)
+
+        self.copy_idata.append((tg_admin_users, "AdminUserRecoveryInfo.plist"))
+
+        admin_users = plistlib.load(open(tg_admin_users, "rb"))
+        self.stub_info["admin_users"] = {}
+        for user, info in admin_users.items():
+            self.stub_info["admin_users"][user] = {
+                "uid": info["GeneratedUID"],
+                "real_name": info["RealName"],
+            }
+
+        # Stop macOS <12.0 bootability stufff from clobbering this file
+        self.chflags("schg", tg_admin_users)
+
+    def repair(self, cur_os):
+        self.get_paths()
+        self.copy_admin_users(cur_os)
+
+        if self.osi.attached_partitions:
+            logging.info("Found attached partitions")
+            efi_part = self.osi.attached_partitions[0]
+            if efi_part.type != "EFI":
+                logging.info("EFI partition not found")
+            else:
+                logging.info("EFI partition found")
+                mountpoint = self.dutil.mount(efi_part.name)
+                asahi = os.path.join(mountpoint, "asahi")
+                if os.path.exists(asahi):
+                    self.collect_installer_data(asahi, merge_stub_info=True)
+                    logging.info("Admin users updated in ESP")
+                else:
+                    logging.info("'asahi' dir not found in ESP")
+
     def install_files(self, cur_os):
         logging.info("StubInstaller.install_files()")
         logging.info(f"VGID: {self.osi.vgid}")
         logging.info(f"OS info: {self.osi}")
 
         p_progress("Beginning stub OS install...")
-        ipsw = self.pkg
-
         self.get_paths()
 
         logging.info("Parsing metadata...")
 
-        sysver = plistlib.load(ipsw.open("SystemVersion.plist"))
-        manifest = plistlib.load(ipsw.open("BuildManifest.plist"))
-        bootcaches = plistlib.load(ipsw.open("usr/standalone/bootcaches.plist"))
+        sysver = plistlib.load(self.open("SystemVersion.plist"))
+        manifest = plistlib.load(self.open("BuildManifest.plist"))
+        bootcaches = plistlib.load(self.open("usr/standalone/bootcaches.plist"))
         self.flush_progress()
 
+        if self.is_ota:
+            variant = "macOS Customer Software Update"
+            behavior = "Update"
+        else:
+            variant = "macOS Customer"
+            behavior = "Erase"
+
         self.manifest = manifest
         for identity in manifest["BuildIdentities"]:
             if (identity["ApBoardID"] != f'0x{self.sysinfo.board_id:02X}' or
                 identity["ApChipID"] != f'0x{self.sysinfo.chip_id:04X}' or
                 identity["Info"]["DeviceClass"] != self.sysinfo.device_class or
-                identity["Info"]["RestoreBehavior"] != "Erase" or
-                identity["Info"]["Variant"] != "macOS Customer"):
+                identity["Info"]["RestoreBehavior"] != behavior or
+                identity["Info"]["Variant"] != variant):
                 continue
             break
         else:
@@ -222,12 +290,11 @@ class StubInstaller(PackageInstaller):
         p_progress("Setting up Preboot volume...")
         logging.info("Setting up Preboot volume")
 
-        pb_vgid = os.path.join(self.osi.preboot, self.osi.vgid)
-        os.makedirs(pb_vgid, exist_ok=True)
+        os.makedirs(self.pb_vgid, exist_ok=True)
 
         bless2 = bootcaches["bless2"]
 
-        restore_bundle = os.path.join(pb_vgid, bless2["RestoreBundlePath"])
+        restore_bundle = os.path.join(self.pb_vgid, bless2["RestoreBundlePath"])
         os.makedirs(restore_bundle, exist_ok=True)
         restore_manifest = os.path.join(restore_bundle, "BuildManifest.plist")
         with open(restore_manifest, "wb") as fd:
@@ -244,12 +311,13 @@ class StubInstaller(PackageInstaller):
         self.extract_file("BootabilityBundle/Restore/Firmware/Bootability.dmg.trustcache",
                           os.path.join(restore_bundle, "Bootability/Bootability.trustcache"))
 
-        self.extract_tree("Firmware/Manifests/restore/macOS Customer/", restore_bundle)
+        self.extract_tree(f"Firmware/Manifests/restore/{variant}/", restore_bundle)
 
         copied = set()
         self.kernel_path = None
         for key, val in identity["Manifest"].items():
-            if key in ("BaseSystem", "OS", "Ap,SystemVolumeCanonicalMetadata"):
+            if key in ("BaseSystem", "OS", "Ap,SystemVolumeCanonicalMetadata",
+                       "RestoreRamDisk", "RestoreTrustCache"):
                 continue
             if key.startswith("Cryptex"):
                 continue
@@ -266,25 +334,7 @@ class StubInstaller(PackageInstaller):
 
         self.flush_progress()
 
-        os.makedirs(os.path.join(pb_vgid, "var/db"), exist_ok=True)
-        admin_users = os.path.join(cur_os.preboot, cur_os.vgid, "var/db/AdminUserRecoveryInfo.plist")
-        tg_admin_users = os.path.join(pb_vgid, "var/db/AdminUserRecoveryInfo.plist")
-        if os.path.exists(tg_admin_users):
-            self.chflags("noschg", tg_admin_users)
-        shutil.copy(admin_users, tg_admin_users)
-
-        self.copy_idata.append((tg_admin_users, "AdminUserRecoveryInfo.plist"))
-
-        admin_users = plistlib.load(open(tg_admin_users, "rb"))
-        self.stub_info["admin_users"] = {}
-        for user, info in admin_users.items():
-            self.stub_info["admin_users"][user] = {
-                "uid": info["GeneratedUID"],
-                "real_name": info["RealName"],
-            }
-
-        # Stop macOS <12.0 bootability stufff from clobbering this file
-        self.chflags("schg", tg_admin_users)
+        self.copy_admin_users(cur_os)
 
         # This is a workaround for some screwiness in the macOS <12.0 bootability
         # code, which ends up putting the apticket in the wrong volume...
@@ -303,8 +353,12 @@ class StubInstaller(PackageInstaller):
         os.makedirs(basesystem_path, exist_ok=True)
 
         logging.info("Extracting arm64eBaseSystem.dmg")
-        self.copy_compress(identity["Manifest"]["BaseSystem"]["Info"]["Path"],
-                          os.path.join(basesystem_path, "arm64eBaseSystem.dmg"))
+        if self.is_ota:
+            self.copy_recompress("AssetData/payloadv2/basesystem_patches/arm64eBaseSystem.dmg",
+                                 os.path.join(basesystem_path, "arm64eBaseSystem.dmg"))
+        else:
+            self.copy_compress(identity["Manifest"]["BaseSystem"]["Info"]["Path"],
+                               os.path.join(basesystem_path, "arm64eBaseSystem.dmg"))
         self.flush_progress()
 
         p_progress("Wrapping up...")
@@ -323,6 +377,7 @@ class StubInstaller(PackageInstaller):
 
         logging.info("Writing step2.sh")
         step2_sh = open("step2/step2.sh").read()
+        step2_sh = step2_sh.replace("##DISTRO##", DISTRO)
         step2_sh = step2_sh.replace("##VGID##", self.osi.vgid)
         step2_sh = step2_sh.replace("##PREBOOT##", self.osi.preboot_vgid)
         with open(self.step2_sh, "w") as fd:
@@ -407,12 +462,19 @@ class StubInstaller(PackageInstaller):
         logging.info("Detaching recovery ramdisk")
         subprocess.run(["hdiutil", "detach", "-quiet", "recovery"])
 
-    def collect_installer_data(self, path):
+    def collect_installer_data(self, path, merge_stub_info=False):
         p_progress("Collecting installer data...")
         logging.info(f"Copying installer data to {path}")
 
         for src, name in self.copy_idata:
             shutil.copy(src, os.path.join(path, name))
 
+        if merge_stub_info:
+            with open(os.path.join(path, "stub_info.json"), "r") as fd:
+                stub_info = json.load(fd)
+            stub_info.update(self.stub_info)
+        else:
+            stub_info = self.stub_info
+
         with open(os.path.join(path, "stub_info.json"), "w") as fd:
-            json.dump(self.stub_info, fd)
+            json.dump(stub_info, fd)
